"""
Load local .env from CWD or path, if provided

Current format for the `.env` file supports strings only and is parsed in
the following order:
- Each seperate line is considered a new possible key/value set
- Each set is delimted by the first `=` found
- Leading and trailing whitespace are removed
- Matched leading/trailing single quotes or double quotes will be
  stripped from values (not keys).

I'm open to suggestions on standards to follow here.

Author  : Preocts <Preocts#8196>
Git Repo: https://github.com/Preocts/secretbox
"""
from __future__ import annotations

import logging
import re

from secretbox.loader import Loader


class EnvFileLoader(Loader):
    """Load local .env file"""

    RE_LTQUOTES = re.compile(r"([\"'])(.*)\1$|^(.*)$")
    EXPORT_PREFIX = r"^\s*?export\s"

    logger = logging.getLogger(__name__)

    def __init__(self, filename: str | None = None) -> None:
        """
        Load local .env file.

        Args:
            filename: Optional filename (with path) to load, default is `.env`
        """
        self._loaded_values: dict[str, str] = {}
        self._filename = filename

    @property
    def values(self) -> dict[str, str]:
        """Copy of loaded values"""
        return self._loaded_values.copy()

    def run(self) -> bool:
        """Load .env, or instantiated filename, to class state."""
        was_loaded = self._load_values()

        for key, value in self._loaded_values.items():
            self.logger.debug("Found, %s : ***%s", key, value[-(len(value) // 4) :])

        return was_loaded

    def _load_values(self, filename: str | None = None, **kwargs: str) -> bool:
        """
        Load values from .env, or provided filename, to class state.

        Args:
            filename : [str] Alternate filename to load over `.env`
        """
        filename = self._filename or filename or ".env"
        self.logger.debug("Reading vars from '%s'", filename)
        try:
            with open(filename, encoding="utf-8") as input_file:
                self.parse_env_file(input_file.read())
        except FileNotFoundError:
            return False
        return True

    def parse_env_file(self, input_file: str) -> None:
        """Parses env file into key-pair values"""
        for line in input_file.split("\n"):
            if not line or line.strip().startswith("#") or len(line.split("=", 1)) != 2:
                continue
            key, value = line.split("=", 1)

            key = self.strip_export(key).strip()
            value = value.strip()

            value = self.remove_lt_quotes(value)

            self._loaded_values[key] = value

    def remove_lt_quotes(self, in_: str) -> str:
        """Removes matched leading and trailing single / double quotes"""
        m = self.RE_LTQUOTES.match(in_)
        return m.group(2) if m and m.group(2) else in_

    def strip_export(self, in_: str) -> str:
        """Removes leading 'export ' prefix, case agnostic"""
        return re.sub(self.EXPORT_PREFIX, "", in_, flags=re.IGNORECASE)



from __future__ import annotations

from abc import ABC
from abc import abstractmethod
from typing import Any


class Loader(ABC):
    """Abstract Base Class for all loaders"""

    @property
    @abstractmethod
    def values(self) -> dict[str, str]:
        """Property: loaded values."""
        raise NotImplementedError()

    @abstractmethod
    def run(self) -> bool:
        """Load values from target source."""
        raise NotImplementedError()

    @abstractmethod
    def _load_values(self, **kwargs: Any) -> bool:
        """Load from source, store values with class instance."""
        raise NotImplementedError()



from __future__ import annotations

from .awsparameterstore_loader import AWSParameterStoreLoader
from .awssecret_loader import AWSSecretLoader
from .envfile_loader import EnvFileLoader
from .environ_loader import EnvironLoader
from .secretbox import SecretBox

__all__ = [
    "AWSParameterStoreLoader",
    "AWSSecretLoader",
    "EnvFileLoader",
    "EnvironLoader",
    "SecretBox",
]



"""
Load system environ values

Author  : Preocts <Preocts#8196>
Git Repo: https://github.com/Preocts/secretbox
"""
from __future__ import annotations

import logging
import os

from secretbox.loader import Loader


class EnvironLoader(Loader):
    """Load environ values"""

    logger = logging.getLogger(__name__)

    def __init__(self) -> None:
        """Load system environ values"""
        self._loaded_values: dict[str, str] = {}

    @property
    def values(self) -> dict[str, str]:
        """Copy of loaded values"""
        return self._loaded_values.copy()

    def _load_values(self, **kwargs: str) -> bool:
        """Load all environmental variables."""
        self.logger.debug("Reading %s environ variables", len(os.environ))
        self._loaded_values.update(os.environ)
        return True

    def run(self) -> bool:
        """Load all environ variables."""
        has_loaded = self._load_values()

        for key, value in self._loaded_values.items():
            self.logger.debug("Found, %s : ***%s", key, value[-(len(value) // 4) :])

        return has_loaded



"""
Load secrets from an AWS secret manager

Author  : Preocts <Preocts#8196>
Git Repo: https://github.com/Preocts/secretbox
"""
from __future__ import annotations

import json
from typing import Any

try:
    import boto3
    from botocore.exceptions import ClientError
    from botocore.exceptions import NoCredentialsError
except ImportError:
    boto3 = None  # type: ignore

try:
    from mypy_boto3_secretsmanager.client import SecretsManagerClient
except ImportError:
    SecretsManagerClient = None  # type: ignore

from secretbox.aws_loader import AWSLoader


class AWSSecretLoader(AWSLoader):
    def __init__(
        self,
        aws_sstore_name: str | None = None,
        aws_region_name: str | None = None,
    ) -> None:
        """
        Load secrets from an AWS secret manager.

        Args:
            aws_sstore: Name of the secret store (not the arn)
                Can be provided through environ `AWS_SSTORE_NAME`
            aws_region: Regional location of secret store
                Can be provided through environ `AWS_REGION_NAME` or `AWS_REGION`
        """
        self.aws_sstore = aws_sstore_name
        self.aws_region = aws_region_name

        self._loaded_values: dict[str, str] = {}

    @property
    def values(self) -> dict[str, str]:
        """Copy of loaded values"""
        return self._loaded_values.copy()

    def run(self) -> bool:
        """Load all secrets from given AWS secret store."""
        has_loaded = self._load_values()

        for key, value in self._loaded_values.items():
            self.logger.debug("Found, %s : ***%s", key, value[-(len(value) // 4) :])

        return has_loaded

    def _load_values(
        self,
        aws_sstore_name: str | None = None,
        aws_region_name: str | None = None,
        **kwargs: Any,
    ) -> bool:
        """
        Load all secrets from AWS secret store.

        Parameter values set at class instantiation are used if values
        are not provided here.

        Args:
            aws_sstore: Name of the secret store (not the arn)
                Can be provided through environ `AWS_SSTORE_NAME`
            aws_region: Regional location of secret store
                Can be provided through environ `AWS_REGION_NAME` or `AWS_REGION`

        Keyword Args:
            Deprecated.
        """
        if boto3 is None:
            self.logger.error("Required boto3 modules missing, can't load AWS secrets")
            return False

        aws_sstore = aws_sstore_name or self.aws_sstore
        aws_region = aws_region_name or self.aws_region

        self.populate_region_store_names(aws_sstore, aws_region, **kwargs)
        if self.aws_sstore is None:
            self.logger.error("Missing secret store name")
            return False

        aws_client = self.get_aws_client()
        if aws_client is None:
            self.logger.error("Invalid secrets manager client")
            return False

        secrets: dict[str, str] = {}
        try:
            # ensure that boto3 doesn't write sensitive payload to the logger
            with self.disable_debug_logging():
                response = aws_client.get_secret_value(SecretId=self.aws_sstore)

        except NoCredentialsError as err:
            self.logger.error("Missing AWS credentials (%s)", err)

        except ClientError as err:
            self.log_aws_error(err)

        else:
            secrets = self._resolve_response(response)
            self._loaded_values.update(secrets)

        return bool(secrets)

    def _resolve_response(self, response: Any) -> dict[str, str]:
        """Resolve response body to json."""
        try:
            secrets = json.loads(response.get("SecretString", "{}"))
        except json.JSONDecodeError:
            secrets = {response.get("Name", ""): response.get("SecretString", "")}
        self.logger.debug("Found %s values from AWS.", len(secrets))

        return secrets

    def get_aws_client(self) -> SecretsManagerClient | None:
        """Return Secrets Manager client"""

        if not self.aws_region:
            self.logger.error("No valid AWS region, cannot create client.")
            return None

        with self.disable_debug_logging():
            client = boto3.client(
                service_name="secretsmanager",
                region_name=self.aws_region,
            )

        return client



"""
Loads various environment variables/secrets for use

Author  : Preocts <Preocts#8196>
Git Repo: https://github.com/Preocts/secretbox
"""
from __future__ import annotations

import logging
import os
from typing import Any

from secretbox.awsparameterstore_loader import (
    AWSParameterStoreLoader as _AWSParameterStoreLoader,
)
from secretbox.awssecret_loader import AWSSecretLoader as _AWSSecretLoader
from secretbox.envfile_loader import EnvFileLoader as _EnvFileLoader
from secretbox.environ_loader import EnvironLoader as _EnvironLoader
from secretbox.loader import Loader

# To be removed with 2.8.0
LOADERS: dict[str, type[Loader]] = {
    "envfile": _EnvFileLoader,
    "environ": _EnvironLoader,
    "awssecret": _AWSSecretLoader,
    "awsparameterstore": _AWSParameterStoreLoader,
}


class SecretBox:
    """Loads various environment variables/secrets for use"""

    _logger = logging.getLogger(__name__)
    AWSParameterStoreLoader = _AWSParameterStoreLoader
    AWSSecretLoader = _AWSSecretLoader
    EnvFileLoader = _EnvFileLoader
    EnvironLoader = _EnvironLoader

    def __init__(
        self,
        *,
        auto_load: bool = False,
        debug_flag: bool = False,
    ) -> None:
        """
        Initialize SecretBox

        Keyword Args:
            auto_load : If true, environment vars and `.env` file will be loaded
            load_debug : When true, internal logger level is set to DEBUG
        """
        self._logger.setLevel(level="DEBUG" if debug_flag else "ERROR")
        self._logger.debug("Debug flag passed.")

        self._loaded_values: dict[str, str] = {}

        if auto_load:
            self.use_loaders(self.EnvironLoader(), self.EnvFileLoader())

    @property
    def values(self) -> dict[str, str]:
        """Property: loaded values."""
        return self._loaded_values.copy()

    def use_loaders(self, *loaders: Loader) -> None:
        """
        Loaded results are injected into environ and stored in state.

        Args:
            loaders: Variable length argument list of Loaders to execute.
        """
        for loader in loaders:
            loader.run()
            self._loaded_values.update(loader.values)

        self._push_to_environment()

    def load_from(
        self,
        loaders: list[str],
        **kwargs: Any,
    ) -> None:
        """
        Runs load_values from each of the listed loader in the order they appear

        Deprecated: This method will be replaced with `.use_loaders()` in v2.7.0

        Loader options:
            environ:
                Loads the current environmental variables into secretbox.
            envfile:
                Loads .env file. Optional `filename` kwarg can override the default
                load of the current working directory `.env` file.
            awssecret:
                Loads secrets from an AWS secret manager. Requires `aws_sstore_name`
                and `aws_region_name` keywords to be provided or for those values
                to be in the environment variables under `AWS_SSTORE_NAME` and
                `AWS_REGION_NAME`. `aws_sstore_name` is not the arn.
        """
        self._logger.warning("Deprecated: `.load_from()` will be removed in v2.8.0")
        for loader_name in loaders:
            self._logger.debug("Loading from interface: `%s`", loader_name)
            interface = LOADERS.get(loader_name)
            if interface is None:
                self._logger.error("Loader `%s` unknown, skipping", loader_name)
                continue
            loader = interface()
            loader._load_values(**kwargs)
            self._logger.debug("Loaded %d values.", len(loader.values))
            self._update_loaded_values(loader.values)
        self._push_to_environment()

    def _update_loaded_values(self, new_values: dict[str, str]) -> None:
        """Update/Create instance state of loaded values with new values"""
        self._loaded_values.update(new_values)

    def _push_to_environment(self) -> None:
        """Pushes loaded values to local environment vars, will overwrite existing"""
        for key, value in self._loaded_values.items():
            self._logger.debug("Push, %s : ***%s", key, value[-(len(value) // 4) :])
            os.environ[key] = value

    def get(self, key: str, default: str | None = None) -> str:
        """Get a value by key, return default if not found or raise if no default"""
        if default is None:
            return self._loaded_values[key]

        return self._loaded_values.get(key, default)

    def set(self, key: str, value: str) -> None:  # noqa: A003
        """Set a value by key. Will be converted to string and pushed to environment."""
        value = str(value)
        self._loaded_values[key] = value
        self._push_to_environment()



"""
Super class for AWS secrets manager and parameter store loaders

Author  : Preocts <preocts#8196>
Git Repo: https://github.com/Preocts/secretbox
"""
from __future__ import annotations

import logging
import os
from contextlib import contextmanager
from typing import Any
from typing import Generator

try:
    from botocore.awsrequest import HeadersDict
except ImportError:
    HeadersDict = dict  # type: ignore

from secretbox.loader import Loader


class AWSLoader(Loader):
    """Super class with mutual methods of AWS loaders, inherits Loader"""

    logger = logging.getLogger(__name__)

    # Override hide_boto_debug to False to allow full debug logging of boto3 libraries
    # NOTE: This exposes sensitive data and should never be in production
    hide_boto_debug = True

    def _load_values(self, **kwargs: Any) -> bool:
        """To be overrided in child classes"""
        raise NotImplementedError()

    @property
    def values(self) -> dict[str, str]:
        """To be overrided in child classes"""
        raise NotImplementedError()

    def run(self) -> bool:
        """To be overrided in child classes"""
        raise NotImplementedError()

    def get_aws_client(self) -> Any:
        """Returns correct AWS client for low-level API requests"""
        # NOTE: Override in client specific implementation
        raise NotImplementedError

    def populate_region_store_names(
        self,
        sstore: str | None = None,
        region: str | None = None,
        **kwargs: str,
    ) -> None:
        """Populate store/region values."""
        kw_sstore = sstore or kwargs.get("aws_sstore_name")
        kw_region = region or kwargs.get("aws_region_name")
        os_sstore = os.getenv("AWS_SSTORE_NAME")
        os_region = os.getenv("AWS_REGION_NAME", os.getenv("AWS_REGION"))  # Lambda's

        # Use the keyword over the os, default to None
        self.aws_sstore = kw_sstore if kw_sstore is not None else os_sstore
        self.aws_region = kw_region if kw_region is not None else os_region
        self.logger.debug("Using store name '%s'", self.aws_sstore)
        self.logger.debug("Using region name '%s'", self.aws_region)

    def log_aws_error(self, err: Any) -> None:
        """Verbose AWS error log output. If not an AWS error, generic repl of err"""
        if (
            hasattr(err, "response")  # assert attribute exists or fail early
            and "Error" in err.response
            and "ResponseMetadata" in err.response
        ):
            self.logger.error(
                "%s - %s (%s)",
                err.response["Error"]["Code"],
                err.response["Error"]["Message"],
                err.response["ResponseMetadata"],
            )
        else:
            self.logger.error("Unexpected error occurred: '%s'", str(err))

    @contextmanager
    def disable_debug_logging(self) -> Generator[None, None, None]:
        """Context manager to enforce level 20 (INFO) logging minimum at root logger."""
        # This should prevent boto and botocore loggers from outputting
        # client secrets in plaintext if debug logging is on.
        prior_root_level = logging.getLogger().level

        if self.hide_boto_debug:
            self.logger.info("Forcing all loggers to > DEBUG level.")
            logging.getLogger().setLevel(logging.INFO)

        try:
            yield None

        finally:
            if self.hide_boto_debug:
                self.logger.info("Restoring previous loggers settings.")
                logging.getLogger().setLevel(prior_root_level)



"""
Author      : James McKain (@jjmckain)
Created     : 2021-12-10
SCM Repo    : https://github.com/Preocts/secretbox
"""
from __future__ import annotations

from typing import Any

from secretbox.aws_loader import AWSLoader

try:
    import boto3
    from botocore.exceptions import ClientError
except ImportError:
    boto3 = None  # type: ignore

try:
    from mypy_boto3_ssm.client import SSMClient
except ImportError:
    SSMClient = None  # type: ignore


class AWSParameterStoreLoader(AWSLoader):
    """Load secrets from an AWS Parameter Store"""

    def __init__(
        self,
        aws_sstore_name: str | None = None,
        aws_region_name: str | None = None,
    ) -> None:
        """
        Load secrets from AWS parameter store.

        Args:
            aws_sstore: Name of parameter or path of parameters if endings with `/`
                Can be provided through environ `AWS_SSTORE_NAME`
            aws_region: Regional Location of parameter(s)
                Can be provided through environ `AWS_REGION_NAME` or `AWS_REGION`
        """
        self.aws_sstore = aws_sstore_name
        self.aws_region = aws_region_name

        self._loaded_values: dict[str, str] = {}

    @property
    def values(self) -> dict[str, str]:
        """Copy of loaded values"""
        return self._loaded_values.copy()

    def run(self) -> bool:
        """Load secrets from given AWS parameter store."""
        has_loaded = self._load_values()

        for key, value in self._loaded_values.items():
            self.logger.debug("Found, %s : ***%s", key, value[-(len(value) // 4) :])

        return has_loaded

    def _load_values(
        self,
        aws_sstore_name: str | None = None,
        aws_region_name: str | None = None,
        **kwargs: Any,
    ) -> bool:
        """
        Load all secrets from AWS parameter store.

        Parameter values set at class instantiation are used if values
        are not provided here.

        Args:
            aws_sstore: Name of parameter or path of parameters if endings with `/`
                Can be provided through environ `AWS_SSTORE_NAME`
            aws_region: Regional Location of parameter(s)
                Can be provided through environ `AWS_REGION_NAME` or `AWS_REGION`

        Keyword Args:
            Deprecated.
        """
        if boto3 is None:
            self.logger.debug("Skipping AWS loader, boto3 is not available.")
            return False

        aws_sstore = aws_sstore_name or self.aws_sstore
        aws_region = aws_region_name or self.aws_region

        self.populate_region_store_names(aws_sstore, aws_region, **kwargs)
        if self.aws_sstore is None:
            self.logger.warning("Missing parameter name")
            return True  # this isn't a failure on our part

        aws_client = self.get_aws_client()
        if aws_client is None:
            self.logger.error("Invalid SSM client")
            return False

        # if the prefix contains forward slashes treat the last token as the key name
        do_split = "/" in self.aws_sstore

        args: dict[str, Any] = {
            "Path": self.aws_sstore,
            "Recursive": True,
            "MaxResults": 10,
            "WithDecryption": True,
        }

        # ensure that boto3 doesn't write sensitive payload to the logger
        with self.disable_debug_logging():
            # loop through next page tokens, page size caps at 10
            while True:

                try:
                    resp = aws_client.get_parameters_by_path(**args)

                except ClientError as err:
                    self.log_aws_error(err)
                    return False

                # Process results, break if finished
                for param in resp["Parameters"] or []:
                    # remove the prefix
                    # we want /path/to/DB_PASSWORD to populate os.env.DB_PASSWORD
                    key = param["Name"].split("/")[-1] if do_split else param["Name"]
                    self._loaded_values[key] = param["Value"]

                args["NextToken"] = resp.get("NextToken")

                if not args["NextToken"]:
                    break

                self.logger.debug("fetching next page: %s", args["NextToken"])

        self.logger.info(
            "loaded %d parameters matching %s",
            len(self._loaded_values),
            self.aws_sstore,
        )
        return True

    def get_aws_client(self) -> SSMClient | None:
        """Make the connection"""

        if self.aws_region is None:
            self.logger.debug("Missing AWS region, cannot create client")
            return None

        with self.disable_debug_logging():
            client = boto3.client(
                service_name="ssm",
                region_name=self.aws_region,
            )

        return client
